# sage_setup: distribution = sagemath-categories
"""nodoctest
all.py -- export of projective schemes to Sage
"""

# ****************************************************************************
#  Copyright (C) 2005 William Stein <wstein@gmail.com>
#  Copyright (C) 2025 Gareth Ma <grhkm21@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.schemes.weighted_projective.weighted_projective_space import WeightedProjectiveSpace
